import os
import pytest
import unittest

from betamax.serializers import BaseSerializer, JSONSerializer


class TestJSONSerializer(unittest.TestCase):
    def setUp(self):
        self.cassette_dir = 'fake_dir'
        self.cassette_name = 'cassette_name'

    def test_generate_cassette_name(self):
        assert (os.path.join('fake_dir', 'cassette_name.json') ==
                JSONSerializer.generate_cassette_name(self.cassette_dir,
                                                      self.cassette_name))

    def test_generate_cassette_name_with_instance(self):
        serializer = JSONSerializer()
        assert (os.path.join('fake_dir', 'cassette_name.json') ==
                serializer.generate_cassette_name(self.cassette_dir,
                                                  self.cassette_name))


class Serializer(BaseSerializer):
    name = 'test'


class TestBaseSerializer(unittest.TestCase):
    def test_serialize_is_an_interface(self):
        serializer = Serializer()
        with pytest.raises(NotImplementedError):
            serializer.serialize({})

    def test_deserialize_is_an_interface(self):
        serializer = Serializer()
        with pytest.raises(NotImplementedError):
            serializer.deserialize('path')

    def test_requires_a_name(self):
        with pytest.raises(ValueError):
            BaseSerializer()



# -*- coding: utf-8 -*-
from .base import BaseSerializer
from betamax.exceptions import MissingDirectoryError

import os


class SerializerProxy(BaseSerializer):

    """
    This is an internal implementation detail of the betamax library.

    No users implementing a serializer should be using this. Developers
    working on betamax need only understand that this handles the logic
    surrounding whether a cassette should be updated, overwritten, or created.

    It provides one consistent way for betamax to be confident in how it
    serializes the data it receives. It allows authors of Serializer classes
    to not have to duplicate how files are handled. It delegates the
    responsibility of actually serializing the data to those classes and
    handles the rest.

    """

    def __init__(self, serializer, cassette_path, allow_serialization=False):
        self.proxied_serializer = serializer
        self.allow_serialization = allow_serialization
        self.cassette_path = cassette_path

    def _ensure_path_exists(self):
        directory, _ = os.path.split(self.cassette_path)
        if not (directory == '' or os.path.isdir(directory)):
            raise MissingDirectoryError(
                'Configured cassette directory \'{0}\' does not exist - try '
                'creating it'.format(directory)
                )
        if not os.path.exists(self.cassette_path):
            open(self.cassette_path, 'w+').close()

    @classmethod
    def find(cls, serialize_with, cassette_library_dir, cassette_name):
        from . import serializer_registry
        serializer = serializer_registry.get(serialize_with)
        if serializer is None:
            raise ValueError(
                'No serializer registered for {0}'.format(serialize_with)
                )

        cassette_path = cls.generate_cassette_name(
            serializer, cassette_library_dir, cassette_name
            )
        return cls(serializer, cassette_path)

    @staticmethod
    def generate_cassette_name(serializer, cassette_library_dir,
                               cassette_name):
        return serializer.generate_cassette_name(
            cassette_library_dir, cassette_name
            )

    def serialize(self, cassette_data):
        if not self.allow_serialization:
            return

        self._ensure_path_exists()

        with open(self.cassette_path, 'w') as fd:
            fd.write(self.proxied_serializer.serialize(cassette_data))

    def deserialize(self):
        self._ensure_path_exists()

        data = {}
        with open(self.cassette_path) as fd:
            data = self.proxied_serializer.deserialize(fd.read())

        return data



import unittest

from betamax import Betamax, matchers
from betamax.adapter import BetamaxAdapter
from betamax.cassette import Cassette
from requests import Session
from requests.adapters import HTTPAdapter


class TestBetamax(unittest.TestCase):
    def setUp(self):
        self.session = Session()
        self.vcr = Betamax(self.session)

    def test_initialization_does_alter_the_session(self):
        for v in self.session.adapters.values():
            assert not isinstance(v, BetamaxAdapter)
            assert isinstance(v, HTTPAdapter)

    def test_entering_context_alters_adapters(self):
        with self.vcr:
            for v in self.session.adapters.values():
                assert isinstance(v, BetamaxAdapter)

    def test_exiting_resets_the_adapters(self):
        with self.vcr:
            pass
        for v in self.session.adapters.values():
            assert not isinstance(v, BetamaxAdapter)

    def test_current_cassette(self):
        assert self.vcr.current_cassette is None
        self.vcr.use_cassette('test')
        assert isinstance(self.vcr.current_cassette, Cassette)

    def test_use_cassette_returns_cassette_object(self):
        assert self.vcr.use_cassette('test') is self.vcr

    def test_register_request_matcher(self):
        class FakeMatcher(object):
            name = 'fake'

        Betamax.register_request_matcher(FakeMatcher)
        assert 'fake' in matchers.matcher_registry
        assert isinstance(matchers.matcher_registry['fake'], FakeMatcher)

    def test_stores_the_session_instance(self):
        assert self.session is self.vcr.session

    def test_replaces_all_adapters(self):
        mount_point = 'fake_protocol://'
        s = Session()
        s.mount(mount_point, HTTPAdapter())
        with Betamax(s):
            adapter = s.adapters.get(mount_point)
            assert adapter is not None
            assert isinstance(adapter, BetamaxAdapter)



from .helper import IntegrationHelper
from betamax import Betamax
from betamax.cassette import Cassette

import copy


class TestPreserveExactBodyBytes(IntegrationHelper):
    def test_preserve_exact_body_bytes_does_not_munge_response_content(self):
        # Do not delete this cassette after the test
        self.cassette_created = False

        with Betamax(self.session) as b:
            b.use_cassette('preserve_exact_bytes',
                           preserve_exact_body_bytes=True,
                           match_requests_on=['uri', 'method', 'body'])
            r = self.session.post('https://httpbin.org/post',
                                  data={'a': 1})
            assert 'headers' in r.json()

            interaction = b.current_cassette.interactions[0].data
            assert 'base64_string' in interaction['request']['body']
            assert 'base64_string' in interaction['response']['body']


class TestPreserveExactBodyBytesForAllCassettes(IntegrationHelper):
    def setUp(self):
        super(TestPreserveExactBodyBytesForAllCassettes, self).setUp()
        self.orig = copy.deepcopy(Cassette.default_cassette_options)
        self.cassette_created = False

    def tearDown(self):
        super(TestPreserveExactBodyBytesForAllCassettes, self).tearDown()
        Cassette.default_cassette_options = self.orig

    def test_preserve_exact_body_bytes(self):
        with Betamax.configure() as config:
            config.preserve_exact_body_bytes = True

        with Betamax(self.session) as b:
            b.use_cassette('global_preserve_exact_body_bytes')
            r = self.session.get('https://httpbin.org/get')
            assert 'headers' in r.json()

            interaction = b.current_cassette.interactions[0].data
            assert 'base64_string' in interaction['response']['body']



import unittest

from requests import PreparedRequest
from requests.cookies import RequestsCookieJar
from betamax import matchers


class TestMatchers(unittest.TestCase):
    def setUp(self):
        self.alt_url = ('http://example.com/path/to/end/point?query=string'
                        '&foo=bar')
        self.p = PreparedRequest()
        self.p.body = 'Foo bar'
        self.p.headers = {'User-Agent': 'betamax/test'}
        self.p.url = 'http://example.com/path/to/end/point?query=string'
        self.p.method = 'GET'
        self.p._cookies = RequestsCookieJar()

    def test_matcher_registry_has_body_matcher(self):
        assert 'body' in matchers.matcher_registry

    def test_matcher_registry_has_digest_auth_matcher(self):
        assert 'digest-auth' in matchers.matcher_registry

    def test_matcher_registry_has_headers_matcher(self):
        assert 'headers' in matchers.matcher_registry

    def test_matcher_registry_has_host_matcher(self):
        assert 'host' in matchers.matcher_registry

    def test_matcher_registry_has_method_matcher(self):
        assert 'method' in matchers.matcher_registry

    def test_matcher_registry_has_path_matcher(self):
        assert 'path' in matchers.matcher_registry

    def test_matcher_registry_has_query_matcher(self):
        assert 'query' in matchers.matcher_registry

    def test_matcher_registry_has_uri_matcher(self):
        assert 'uri' in matchers.matcher_registry

    def test_body_matcher(self):
        match = matchers.matcher_registry['body'].match
        assert match(self.p, {
            'body': 'Foo bar',
            'headers': {'User-Agent': 'betamax/test'},
            'uri': 'http://example.com/path/to/end/point?query=string',
            'method': 'GET',
        })
        assert match(self.p, {
            'body': b'',
            'headers': {'User-Agent': 'betamax/test'},
            'uri': 'http://example.com/path/to/end/point?query=string',
            'method': 'GET',
        }) is False

    def test_body_matcher_without_body(self):
        p = self.p.copy()
        p.body = None
        match = matchers.matcher_registry['body'].match
        assert match(p, {
            'body': 'Foo bar',
            'headers': {'User-Agent': 'betamax/test'},
            'uri': 'http://example.com/path/to/end/point?query=string',
            'method': 'GET',
        }) is False
        assert match(p, {
            'body': b'',
            'headers': {'User-Agent': 'betamax/test'},
            'uri': 'http://example.com/path/to/end/point?query=string',
            'method': 'GET',
        })

    def test_digest_matcher(self):
        match = matchers.matcher_registry['digest-auth'].match
        assert match(self.p, {'headers': {}})
        saved_auth = (
            'Digest username="user", realm="realm", nonce="nonce", uri="/", '
            'response="r", opaque="o", qop="auth", nc=00000001, cnonce="c"'
            )
        self.p.headers['Authorization'] = saved_auth
        assert match(self.p, {'headers': {}}) is False
        assert match(self.p, {'headers': {'Authorization': saved_auth}})
        new_auth = (
            'Digest username="user", realm="realm", nonce="nonce", uri="/", '
            'response="e", opaque="o", qop="auth", nc=00000001, cnonce="n"'
            )
        assert match(self.p, {'headers': {'Authorization': new_auth}})
        new_auth = (
            'Digest username="u", realm="realm", nonce="nonce", uri="/", '
            'response="e", opaque="o", qop="auth", nc=00000001, cnonce="n"'
            )
        assert match(self.p, {'headers': {'Authorization': new_auth}}) is False

    def test_headers_matcher(self):
        match = matchers.matcher_registry['headers'].match
        assert match(self.p, {'headers': {'User-Agent': 'betamax/test'}})
        assert match(self.p, {'headers': {'X-Sha': '6bbde0af'}}) is False

    def test_host_matcher(self):
        match = matchers.matcher_registry['host'].match
        assert match(self.p, {'uri': 'http://example.com'})
        assert match(self.p, {'uri': 'https://example.com'})
        assert match(self.p, {'uri': 'https://example.com/path'})
        assert match(self.p, {'uri': 'https://example2.com'}) is False

    def test_method_matcher(self):
        match = matchers.matcher_registry['method'].match
        assert match(self.p, {'method': 'GET'})
        assert match(self.p, {'method': 'POST'}) is False

    def test_path_matcher(self):
        match = matchers.matcher_registry['path'].match
        assert match(self.p, {'uri': 'http://example.com/path/to/end/point'})
        assert match(self.p,
                     {'uri': 'http://example.com:8000/path/to/end/point'})
        assert match(self.p,
                     {'uri': 'http://example.com:8000/path/to/end/'}) is False

    def test_query_matcher(self):
        match = matchers.matcher_registry['query'].match
        assert match(
            self.p,
            {'uri': 'http://example.com/path/to/end/point?query=string'}
        )
        assert match(
            self.p,
            {'uri': 'http://example.com/?query=string'}
        )
        self.p.url = self.alt_url
        assert match(
            self.p,
            {'uri': self.alt_url}
        )
        # Regression test (order independence)
        assert match(
            self.p,
            {'uri': 'http://example.com/?foo=bar&query=string'}
        )
        # Regression test (no query issue)
        assert match(self.p, {'uri': 'http://example.com'}) is False
        # Regression test (query with no value)
        self.p.url = 'https://example.com/?foo'
        assert match(self.p, {'uri': 'https://httpbin.org/?foo'}) is True

    def test_uri_matcher(self):
        match = matchers.matcher_registry['uri'].match
        assert match(
            self.p,
            {'uri': 'http://example.com/path/to/end/point?query=string'}
        )
        assert match(self.p, {'uri': 'http://example.com'}) is False

    def test_uri_matcher_handles_query_strings(self):
        match = matchers.matcher_registry['uri'].match
        self.p.url = 'http://example.com/path/to?query=string&form=value'
        other_uri = 'http://example.com/path/to?form=value&query=string'
        assert match(self.p, {'uri': other_uri}) is True


class TestBaseMatcher(unittest.TestCase):
    def setUp(self):
        class Matcher(matchers.BaseMatcher):
            pass
        self.Matcher = Matcher

    def test_requires_name(self):
        self.assertRaises(ValueError, self.Matcher)

    def test_requires_you_overload_match(self):
        self.Matcher.name = 'test'
        m = self.Matcher()
        self.assertRaises(NotImplementedError, m.match, None, None)



"""
betamax.adapter.

==============

adapter for betamax
"""
import os

from . import cassette
from .exceptions import BetamaxError
from datetime import datetime, timedelta
from requests.adapters import BaseAdapter, HTTPAdapter

_SENTINEL = object()


class BetamaxAdapter(BaseAdapter):
    """This object is an implementation detail of the library.

    It is not meant to be a public API and is not exported as such.

    """

    def __init__(self, **kwargs):
        super(BetamaxAdapter, self).__init__()
        self.cassette = None
        self.cassette_name = None
        self.old_adapters = kwargs.pop('old_adapters', {})
        self.http_adapter = HTTPAdapter(**kwargs)
        self.serialize = None
        self.options = {}

    def cassette_exists(self):
        """Check if cassette exists on file system.

        :returns: bool -- True if exists, False otherwise
        """
        if self.cassette_name and os.path.exists(self.cassette_name):
            return True
        return False

    def close(self):
        """Propagate close to underlying adapter."""
        self.http_adapter.close()

    def eject_cassette(self):
        """Eject currently loaded cassette."""
        if self.cassette:
            self.cassette.eject()
        self.cassette = None  # Allow self.cassette to be garbage-collected

    def load_cassette(self, cassette_name, serialize, options):
        """Load cassette.

        Loads a previously serialized http response as a cassette

        :param str cassette_name: (required), name of cassette
        :param str serialize: (required), type of serialization i.e 'json'
        :options dict options: (required), options for cassette
        """
        self.cassette_name = cassette_name
        self.serialize = serialize
        self.options.update(options.items())
        placeholders = self.options.get('placeholders', {})
        cassette_options = {}

        default_options = cassette.Cassette.default_cassette_options

        match_requests_on = self.options.get(
            'match_requests_on', default_options['match_requests_on']
            )

        cassette_options['preserve_exact_body_bytes'] = self.options.get(
            'preserve_exact_body_bytes',
            )

        cassette_options['allow_playback_repeats'] = self.options.get(
            'allow_playback_repeats'
            )

        cassette_options['record_mode'] = self.options.get('record')

        for option, value in list(cassette_options.items()):
            if value is None:
                cassette_options.pop(option)

        self.cassette = cassette.Cassette(
            cassette_name, serialize, placeholders=placeholders,
            cassette_library_dir=self.options.get('cassette_library_dir'),
            **cassette_options
            )

        if 'record' in self.options:
            self.cassette.record_mode = self.options['record']

        # NOTE(sigmavirus24): Cassette.match_options is a set, might as well
        # use that instead of overriding it.
        self.cassette.match_options.update(match_requests_on)

        re_record_interval = timedelta.max
        if self.options.get('re_record_interval'):
            re_record_interval = timedelta(self.options['re_record_interval'])

        now = datetime.utcnow()
        if re_record_interval < (now - self.cassette.earliest_recorded_date):
            self.cassette.clear()

    def send(self, request, stream=False, timeout=None, verify=True,
             cert=None, proxies=None):
        """Send request.

        :param request request: request
        :returns: A Response object
        """
        interaction = None
        current_cassette = self.cassette

        if not current_cassette:
            raise BetamaxError('No cassette was specified or found.')

        if current_cassette.interactions:
            interaction = current_cassette.find_match(request)

        if not interaction and current_cassette.is_recording():
            interaction = self.send_and_record(
                request, stream, timeout, verify, cert, proxies
                )

        if not interaction:
            raise BetamaxError(unhandled_request_message(request,
                                                         current_cassette))

        resp = interaction.as_response()
        resp.connection = self
        return resp

    def send_and_record(self, request, stream=False, timeout=None,
                        verify=True, cert=None, proxies=None):
        """Send request and record response.

        The response will be serialized and saved to a
        cassette which can be replayed in the future.

        :param request request: request
        :param bool stream: (optional) defer download until content is accessed
        :param float timeout: (optional) time to wait for a response
        :param bool verify: (optional) verify SSL certificate
        :param str cert: (optional) path to SSL client
        :param proxies dict: (optional) mapping protocol to URL of the proxy
        :return: Interaction
        :rtype: class:`betamax.cassette.Interaction`
        """
        adapter = self.find_adapter(request.url)
        response = adapter.send(
            request, stream=True, timeout=timeout, verify=verify,
            cert=cert, proxies=proxies
            )
        return self.cassette.save_interaction(response, request)

    def find_adapter(self, url):
        """Find adapter.

        Searches for an existing adapter where the url and prefix match.

        :param url str: (required) url of the adapter
        :returns: betamax adapter
        """
        for (prefix, adapter) in self.old_adapters.items():

            if url.lower().startswith(prefix):
                return adapter

        # Unlike in requests, we cannot possibly get this far.


UNHANDLED_REQUEST_EXCEPTION = """A request was made that could not be handled.

A request was made to {url} that could not be found in {cassette_file_path}.

The settings on the cassette are:

    - record_mode: {cassette_record_mode}
    - match_options {cassette_match_options}.
"""


def unhandled_request_message(request, cassette):
    """Generate exception for unhandled requests."""
    return UNHANDLED_REQUEST_EXCEPTION.format(
        url=request.url, cassette_file_path=cassette.cassette_name,
        cassette_record_mode=cassette.record_mode,
        cassette_match_options=cassette.match_options
        )



import betamax

from tests.integration import helper


class TestPlaybackRepeatInteractions(helper.IntegrationHelper):
    def test_will_replay_the_same_interaction(self):
        self.cassette_created = False
        s = self.session
        recorder = betamax.Betamax(s)
        # NOTE(sigmavirus24): Ensure the cassette is recorded
        with recorder.use_cassette('replay_interactions'):
            cassette = recorder.current_cassette
            r = s.get('http://httpbin.org/get')
            assert r.status_code == 200
            assert len(cassette.interactions) == 1

        with recorder.use_cassette('replay_interactions',
                                   allow_playback_repeats=True):
            cassette = recorder.current_cassette
            r = s.get('http://httpbin.org/get')
            assert r.status_code == 200
            assert len(cassette.interactions) == 1
            r = s.get('http://httpbin.org/get')
            assert r.status_code == 200
            assert len(cassette.interactions) == 1
            assert cassette.interactions[0].used is False



import os
import unittest

from betamax import Betamax
from requests import Session


class TestGZIPRegression(unittest.TestCase):
    def tearDown(self):
        os.unlink('tests/cassettes/gzip_regression.json')

    def test_saves_content_as_gzip(self):
        s = Session()
        with Betamax(s).use_cassette('gzip_regression'):
            r = s.get(
                'https://api.github.com/repos/github3py/fork_this/issues/1',
                headers={'Accept-Encoding': 'gzip, deflate, compress'}
                )
            assert r.headers.get('Content-Encoding') == 'gzip'
            assert r.json() is not None

            r2 = s.get(
                'https://api.github.com/repos/github3py/fork_this/issues/1',
                headers={'Accept-Encoding': 'gzip, deflate, compress'}
                )
            assert r2.headers.get('Content-Encoding') == 'gzip'
            assert r2.json() is not None
            assert r2.json() == r.json()

        s = Session()
        with Betamax(s).use_cassette('gzip_regression'):
            r = s.get(
                'https://api.github.com/repos/github3py/fork_this/issues/1'
                )
            assert r.json() is not None



from betamax import Betamax
from tests.integration.helper import IntegrationHelper


class TestUnicode(IntegrationHelper):
    def test_unicode_is_saved_properly(self):
        s = self.session
        # https://github.com/kanzure/python-requestions/issues/4
        url = 'http://www.amazon.com/review/RAYTXRF3122TO'

        with Betamax(s).use_cassette('test_unicode') as beta:
            self.cassette_path = beta.current_cassette.cassette_path
            s.get(url)



import betamax

from .helper import IntegrationHelper


class TestMultipleCookies(IntegrationHelper):
    """Previously our handling of multiple instances of cookies was wrong.

    This set of tests is here to ensure that we properly serialize/deserialize
    the case where the client receives and betamax serializes multiple
    Set-Cookie headers.

    See the following for more information:

    - https://github.com/sigmavirus24/betamax/pull/60
    - https://github.com/sigmavirus24/betamax/pull/59
    - https://github.com/sigmavirus24/betamax/issues/58
    """
    def setUp(self):
        super(TestMultipleCookies, self).setUp()
        self.cassette_created = False

    def test_multiple_cookies(self):
        """Make a request to httpbin.org and verify we serialize it correctly.

        We should be able to see that the cookiejar on the session has the
        cookies properly parsed and distinguished.
        """
        recorder = betamax.Betamax(self.session)
        cassette_name = 'test-multiple-cookies-regression'
        url = 'https://httpbin.org/cookies/set'
        cookies = {
            'cookie0': 'value0',
            'cookie1': 'value1',
            'cookie2': 'value2',
            'cookie3': 'value3',
        }
        with recorder.use_cassette(cassette_name):
            self.session.get(url, params=cookies)

        for name, value in cookies.items():
            assert self.session.cookies[name] == value



import os
import unittest

from requests import Session


class IntegrationHelper(unittest.TestCase):
    cassette_created = True

    def setUp(self):
        self.cassette_path = None
        self.session = Session()

    def tearDown(self):
        if self.cassette_created:
            assert self.cassette_path is not None
            os.unlink(self.cassette_path)



# -*- coding: utf-8 -*-
"""A set of fixtures to integrate Betamax with py.test.

.. autofunction:: betamax_session

"""

from __future__ import absolute_import

import pytest
import requests

from .. import recorder as betamax


@pytest.fixture
def betamax_recorder(request):
    """Generate a recorder with a session that has Betamax already installed.

    This will create a new Betamax instance with a generated cassette name.
    The cassette name is generated by first using the module name from where
    the test is collected, then the class name (if it exists), and then the
    test function name. For example, if your test is in ``test_stuff.py`` and
    is the method ``TestStuffClass.test_stuff`` then your cassette name will be
    ``test_stuff_TestStuffClass_test_stuff``.

    :param request:
        A request object from pytest giving us context information for the
        fixture.
    :returns:
        An instantiated recorder.
    """
    cassette_name = ''

    if request.module is not None:
        cassette_name += request.module.__name__ + '.'

    if request.cls is not None:
        cassette_name += request.cls.__name__ + '.'

    cassette_name += request.function.__name__

    session = requests.Session()
    recorder = betamax.Betamax(session)
    recorder.use_cassette(cassette_name)
    recorder.start()
    request.addfinalizer(recorder.stop)
    return recorder


@pytest.fixture
def betamax_session(betamax_recorder):
    """Generate a session that has Betamax already installed.

    See `betamax_recorder` fixture.

    :param betamax_recorder:
        A recorder fixture with a configured request session.
    :returns:
        An instantiated requests Session wrapped by Betamax.
    """

    return betamax_recorder.session



import pytest
import unittest

from betamax import Betamax, BetamaxError
from requests import Session


class TestOncePreventsNewInteractions(unittest.TestCase):

    """Test that using a cassette with once record mode prevents new requests.

    """

    def test_once_prevents_new_requests(self):
        s = Session()
        with Betamax(s).use_cassette('once_record_mode'):
            with pytest.raises(BetamaxError):
                s.get('http://example.com')



import pytest
import unittest

from betamax import Betamax, cassette
from requests import Session


class TestCassetteRecordMode(unittest.TestCase):
    def setUp(self):
        with Betamax.configure() as config:
            config.default_cassette_options['record_mode'] = 'none'

    def tearDown(self):
        with Betamax.configure() as config:
            config.default_cassette_options['record_mode'] = 'once'

    def test_record_mode_is_none(self):
        s = Session()
        with pytest.raises(ValueError):
            with Betamax(s) as recorder:
                recorder.use_cassette('regression_record_mode')
                assert recorder.current_cassette is None

    def test_class_variables_retain_their_value(self):
        opts = cassette.Cassette.default_cassette_options
        assert opts['record_mode'] == 'none'



import functools
import unittest

import requests

from . import recorder


def use_cassette(cassette_name, cassette_library_dir=None,
                 default_cassette_options={}, **use_cassette_kwargs):
    """Provide a Betamax-wrapped Session for convenience.

    .. versionadded:: 0.5.0

    This decorator can be used to get a plain Session that has been wrapped in
    Betamax. For example,

    .. code-block:: python

        from betamax.decorator import use_cassette

        @use_cassette('example-decorator', cassette_library_dir='.')
        def test_get(session):
            # do things with session

    :param str cassette_name:
        Name of the cassette file in which interactions will be stored.
    :param str cassette_library_dir:
        Directory in which cassette files will be stored.
    :param dict default_cassette_options:
        Dictionary of default cassette options to set for the cassette used
        when recording these interactions.
    :param \*\*use_cassette_kwargs:
        Keyword arguments passed to :meth:`~betamax.Betamax.use_cassette`
    """
    def actual_decorator(func):
        @functools.wraps(func)
        def test_wrapper(*args, **kwargs):
            session = requests.Session()
            recr = recorder.Betamax(
                session=session,
                cassette_library_dir=cassette_library_dir,
                default_cassette_options=default_cassette_options
            )

            if args:
                fst, args = args[0], args[1:]
                if isinstance(fst, unittest.TestCase):
                    args = (fst, session) + args
                else:
                    args = (session, fst) + args
            else:
                args = (session,)

            with recr.use_cassette(cassette_name, **use_cassette_kwargs):
                func(*args, **kwargs)

        return test_wrapper
    return actual_decorator



from .base import BaseSerializer

import json
import os


class JSONSerializer(BaseSerializer):
    # Serializes and deserializes a cassette to JSON
    name = 'json'

    @staticmethod
    def generate_cassette_name(cassette_library_dir, cassette_name):
        return os.path.join(cassette_library_dir,
                            '{0}.{1}'.format(cassette_name, 'json'))

    def serialize(self, cassette_data):
        return json.dumps(cassette_data)

    def deserialize(self, cassette_data):
        try:
            deserialized_data = json.loads(cassette_data)
        except ValueError:
            deserialized_data = {}

        return deserialized_data



# -*- coding: utf-8 -*-
from .base import BaseMatcher
from betamax.util import from_list


class DigestAuthMatcher(BaseMatcher):

    """This matcher is provided to help those who need to use Digest Auth.

    .. note::

        The code requests 2.0.1 uses to generate this header is different from
        the code that every requests version after it uses. Specifically, in
        2.0.1 one of the parameters is ``qop=auth`` and every other version is
        ``qop="auth"``. Given that there's also an unsupported type of ``qop``
        in requests, I've chosen not to ignore ore sanitize this. All
        cassettes recorded on 2.0.1 will need to be re-recorded for any
        requests version after it.

        This matcher also ignores the ``cnonce`` and ``response`` parameters.
        These parameters require the system time to be monkey-patched and
        that is out of the scope of betamax

    """

    name = 'digest-auth'

    def match(self, request, recorded_request):
        request_digest = self.digest_parts(request.headers)
        recorded_digest = self.digest_parts(recorded_request['headers'])
        return request_digest == recorded_digest

    def digest_parts(self, headers):
        auth = headers.get('Authorization') or headers.get('authorization')
        if not auth:
            return None
        auth = from_list(auth).strip('Digest ')
        # cnonce and response will be based on the system time, which I will
        # not monkey-patch.
        excludes = ('cnonce', 'response')
        return [p for p in auth.split(', ') if not p.startswith(excludes)]



from betamax import Betamax
from betamax.cassette import Cassette

from copy import deepcopy
from tests.integration.helper import IntegrationHelper

original_cassette_options = deepcopy(Cassette.default_cassette_options)
b64_foobar = 'Zm9vOmJhcg=='  # base64.b64encode('foo:bar')


class TestPlaceholders(IntegrationHelper):
    def setUp(self):
        super(TestPlaceholders, self).setUp()
        config = Betamax.configure()
        config.define_cassette_placeholder('<AUTHORIZATION>', b64_foobar)

    def tearDown(self):
        super(TestPlaceholders, self).tearDown()
        Cassette.default_cassette_options = original_cassette_options

    def test_placeholders_work(self):
        placeholders = Cassette.default_cassette_options['placeholders']
        assert placeholders == [{
            'placeholder': '<AUTHORIZATION>',
            'replace': b64_foobar,
        }]

        s = self.session
        cassette = None
        with Betamax(s).use_cassette('test_placeholders') as recorder:
            r = s.get('http://httpbin.org/get', auth=('foo', 'bar'))
            cassette = recorder.current_cassette
            self.cassette_path = cassette.cassette_path
            assert r.status_code == 200
            auth = r.json()['headers']['Authorization']
            assert b64_foobar in auth

        self.cassette_path = cassette.cassette_path
        i = cassette.interactions[0]
        auth = i.data['request']['headers']['Authorization']
        assert '<AUTHORIZATION>' in auth



import betamax

from . import helper


def prerecord_hook(interaction, cassette):
    assert cassette.interactions == []
    interaction.data['response']['headers']['Betamax-Fake-Header'] = 'success'


def ignoring_hook(interaction, cassette):
    interaction.ignore()


def preplayback_hook(interaction, cassette):
    assert cassette.interactions != []
    interaction.data['response']['headers']['Betamax-Fake-Header'] = 'temp'


class TestHooks(helper.IntegrationHelper):
    def tearDown(self):
        super(TestHooks, self).tearDown()
        # Clear out the hooks
        betamax.cassette.Cassette.hooks.pop('before_record', None)
        betamax.cassette.Cassette.hooks.pop('before_playback', None)

    def test_prerecord_hook(self):
        with betamax.Betamax.configure() as config:
            config.before_record(callback=prerecord_hook)

        recorder = betamax.Betamax(self.session)
        with recorder.use_cassette('prerecord_hook'):
            self.cassette_path = recorder.current_cassette.cassette_path
            response = self.session.get('https://httpbin.org/get')
            assert response.headers['Betamax-Fake-Header'] == 'success'

        with recorder.use_cassette('prerecord_hook', record='none'):
            response = self.session.get('https://httpbin.org/get')
            assert response.headers['Betamax-Fake-Header'] == 'success'

    def test_preplayback_hook(self):
        with betamax.Betamax.configure() as config:
            config.before_playback(callback=preplayback_hook)

        recorder = betamax.Betamax(self.session)
        with recorder.use_cassette('preplayback_hook'):
            self.cassette_path = recorder.current_cassette.cassette_path
            self.session.get('https://httpbin.org/get')

        with recorder.use_cassette('preplayback_hook', record='none'):
            response = self.session.get('https://httpbin.org/get')
            assert response.headers['Betamax-Fake-Header'] == 'temp'

    def test_prerecord_ignoring_hook(self):
        with betamax.Betamax.configure() as config:
            config.before_record(callback=ignoring_hook)

        recorder = betamax.Betamax(self.session)
        with recorder.use_cassette('ignore_hook'):
            self.cassette_path = recorder.current_cassette.cassette_path
            self.session.get('https://httpbin.org/get')
            assert recorder.current_cassette.interactions == []



import email
import os
import unittest
from datetime import datetime

from betamax import __version__
from betamax.cassette import cassette
from betamax import mock_response
from betamax import recorder
from betamax import serializers
from betamax import util
from requests.models import Response, Request
from requests.packages import urllib3
from requests.structures import CaseInsensitiveDict

try:
    from requests.packages.urllib3._collections import HTTPHeaderDict
except ImportError:
    from betamax.headers import HTTPHeaderDict


def decode(s):
    if hasattr(s, 'decode'):
        return s.decode()
    return s


class Serializer(serializers.BaseSerializer):
    name = 'test'

    @staticmethod
    def generate_cassette_name(cassette_library_dir, cassette_name):
        return 'test_cassette.test'

    def on_init(self):
        self.serialize_calls = []
        self.deserialize_calls = []

    def serialize(self, data):
        self.serialize_calls.append(data)
        return ''

    def deserialize(self, data):
        self.deserialize_calls.append(data)
        return {}


class TestSerialization(unittest.TestCase):

    """Unittests for the serialization and deserialization functions.

    This tests:

        - deserialize_prepared_request
        - deserialize_response
        - serialize_prepared_request
        - serialize_response

    """

    def test_serialize_response(self):
        r = Response()
        r.status_code = 200
        r.reason = 'OK'
        r.encoding = 'utf-8'
        r.headers = CaseInsensitiveDict()
        r.url = 'http://example.com'
        util.add_urllib3_response({
            'body': {
                'string': decode('foo'),
                'encoding': 'utf-8'
            }
        }, r, HTTPHeaderDict())
        serialized = util.serialize_response(r, False)
        assert serialized is not None
        assert serialized != {}
        assert serialized['body']['encoding'] == 'utf-8'
        assert serialized['body']['string'] == 'foo'
        assert serialized['headers'] == {}
        assert serialized['url'] == 'http://example.com'
        assert serialized['status'] == {'code': 200, 'message': 'OK'}

    def test_deserialize_response_old(self):
        """For the previous version of Betamax and backwards compatibility."""
        s = {
            'body': {
                'string': decode('foo'),
                'encoding': 'utf-8'
            },
            'headers': {
                'Content-Type': decode('application/json')
            },
            'url': 'http://example.com/',
            'status_code': 200,
            'recorded_at': '2013-08-31T00:00:01'
        }
        r = util.deserialize_response(s)
        assert r.content == b'foo'
        assert r.encoding == 'utf-8'
        assert r.headers == {'Content-Type': 'application/json'}
        assert r.url == 'http://example.com/'
        assert r.status_code == 200

    def test_deserialize_response_new(self):
        """This adheres to the correct cassette specification."""
        s = {
            'body': {
                'string': decode('foo'),
                'encoding': 'utf-8'
            },
            'headers': {
                'Content-Type': [decode('application/json')]
            },
            'url': 'http://example.com/',
            'status': {'code': 200, 'message': 'OK'},
            'recorded_at': '2013-08-31T00:00:01'
        }
        r = util.deserialize_response(s)
        assert r.content == b'foo'
        assert r.encoding == 'utf-8'
        assert r.headers == {'Content-Type': 'application/json'}
        assert r.url == 'http://example.com/'
        assert r.status_code == 200
        assert r.reason == 'OK'

    def test_serialize_prepared_request(self):
        r = Request()
        r.method = 'GET'
        r.url = 'http://example.com'
        r.headers = {'User-Agent': 'betamax/test header'}
        r.data = {'key': 'value'}
        p = r.prepare()
        serialized = util.serialize_prepared_request(p, False)
        assert serialized is not None
        assert serialized != {}
        assert serialized['method'] == 'GET'
        assert serialized['uri'] == 'http://example.com/'
        assert serialized['headers'] == {
            'Content-Length': ['9'],
            'Content-Type': ['application/x-www-form-urlencoded'],
            'User-Agent': ['betamax/test header'],
        }
        assert serialized['body']['string'] == 'key=value'

    def test_deserialize_prepared_request(self):
        s = {
            'body': 'key=value',
            'headers': {
                'User-Agent': 'betamax/test header',
            },
            'method': 'GET',
            'uri': 'http://example.com/',
        }
        p = util.deserialize_prepared_request(s)
        assert p.body == 'key=value'
        assert p.headers == CaseInsensitiveDict(
            {'User-Agent': 'betamax/test header'}
        )
        assert p.method == 'GET'
        assert p.url == 'http://example.com/'

    def test_from_list_returns_an_element(self):
        a = ['value']
        assert util.from_list(a) == 'value'

    def test_from_list_handles_non_lists(self):
        a = 'value'
        assert util.from_list(a) == 'value'

    def test_add_urllib3_response(self):
        r = Response()
        r.status_code = 200
        r.headers = {}
        util.add_urllib3_response({
            'body': {
                'string': decode('foo'),
                'encoding': 'utf-8'
            }
        }, r, HTTPHeaderDict())
        assert isinstance(r.raw, urllib3.response.HTTPResponse)
        assert r.content == b'foo'
        assert isinstance(r.raw._original_response,
                          mock_response.MockHTTPResponse)


def test_cassette_initialization():
    serializers.serializer_registry['test'] = Serializer()
    cassette.Cassette.default_cassette_options['placeholders'] = []

    with recorder.Betamax.configure() as config:
        config.define_cassette_placeholder('<TO-OVERRIDE>', 'default')
        config.define_cassette_placeholder('<KEEP-DEFAULT>', 'config')
        placeholders = [{
            'placeholder': '<TO-OVERRIDE>',
            'replace': 'override',
        }, {
            'placeholder': '<ONLY-OVERRIDE>',
            'replace': 'cassette',
        }]
        instance = cassette.Cassette(
            'test_cassette',
            'test',
            placeholders=placeholders
        )

        expected = [
            cassette.Placeholder('<TO-OVERRIDE>', 'override'),
            cassette.Placeholder('<KEEP-DEFAULT>', 'config'),
            cassette.Placeholder('<ONLY-OVERRIDE>', 'cassette'),
        ]
        assert instance.placeholders == expected

    cassette.Cassette.default_cassette_options['placeholders'] = []


class TestCassette(unittest.TestCase):
    cassette_name = 'test_cassette'

    def setUp(self):
        # Make a new serializer to test with
        self.test_serializer = Serializer()
        serializers.serializer_registry['test'] = self.test_serializer

        # Instantiate the cassette to test with
        self.cassette = cassette.Cassette(
            TestCassette.cassette_name,
            'test',
            record_mode='once'
        )

        # Create a new object to serialize
        r = Response()
        r.status_code = 200
        r.reason = 'OK'
        r.encoding = 'utf-8'
        r.headers = CaseInsensitiveDict({'Content-Type': decode('foo')})
        r.url = 'http://example.com'
        util.add_urllib3_response({
            'body': {
                'string': decode('foo'),
                'encoding': 'utf-8'
            }
        }, r, HTTPHeaderDict({'Content-Type': decode('foo')}))
        self.response = r

        # Create an associated request
        r = Request()
        r.method = 'GET'
        r.url = 'http://example.com'
        r.headers = {}
        r.data = {'key': 'value'}
        self.response.request = r.prepare()
        self.response.request.headers.update(
            {'User-Agent': 'betamax/test header'}
        )

        # Expected serialized cassette data.
        self.json = {
            'request': {
                'body': {
                    'encoding': 'utf-8',
                    'string': 'key=value',
                },
                'headers': {
                    'User-Agent': ['betamax/test header'],
                    'Content-Length': ['9'],
                    'Content-Type': ['application/x-www-form-urlencoded'],
                },
                'method': 'GET',
                'uri': 'http://example.com/',
            },
            'response': {
                'body': {
                    'string': decode('foo'),
                    'encoding': 'utf-8',
                },
                'headers': {'Content-Type': [decode('foo')]},
                'status': {'code': 200, 'message': 'OK'},
                'url': 'http://example.com',
            },
            'recorded_at': '2013-08-31T00:00:00',
        }
        self.date = datetime(2013, 8, 31)
        self.cassette.save_interaction(self.response, self.response.request)
        self.interaction = self.cassette.interactions[0]

    def tearDown(self):
        try:
            self.cassette.eject()
        except:
            pass
        if os.path.exists(TestCassette.cassette_name):
            os.unlink(TestCassette.cassette_name)

    def test_serialize_interaction(self):
        serialized = self.interaction.data
        assert serialized['request'] == self.json['request']
        assert serialized['response'] == self.json['response']
        assert serialized.get('recorded_at') is not None

    def test_holds_interactions(self):
        assert isinstance(self.cassette.interactions, list)
        assert self.cassette.interactions != []
        assert self.interaction in self.cassette.interactions

    def test_find_match(self):
        self.cassette.match_options = set(['uri', 'method'])
        self.cassette.record_mode = 'none'
        i = self.cassette.find_match(self.response.request)
        assert i is not None
        assert self.interaction is i

    def test_eject(self):
        serializer = self.test_serializer
        self.cassette.eject()
        assert serializer.serialize_calls == [
            {'http_interactions': [self.cassette.interactions[0].data],
             'recorded_with': 'betamax/{0}'.format(__version__)}
            ]

    def test_earliest_recorded_date(self):
        assert self.interaction.recorded_at is not None
        assert self.cassette.earliest_recorded_date is not None


class TestInteraction(unittest.TestCase):
    def setUp(self):
        self.request = {
            'body': {
                'string': 'key=value&key2=secret_value',
                'encoding': 'utf-8'
            },
            'headers': {
                'User-Agent': ['betamax/test header'],
                'Content-Length': ['9'],
                'Content-Type': ['application/x-www-form-urlencoded'],
                'Authorization': ['123456789abcdef'],
                },
            'method': 'GET',
            'uri': 'http://example.com/',
        }
        self.response = {
            'body': {
                'string': decode('foo'),
                'encoding': 'utf-8'
            },
            'headers': {
                'Content-Type': [decode('foo')],
                'Set-Cookie': ['cookie_name=cookie_value']
            },
            'status_code': 200,
            'url': 'http://example.com',
        }
        self.json = {
            'request': self.request,
            'response': self.response,
            'recorded_at': '2013-08-31T00:00:00',
        }
        self.interaction = cassette.Interaction(self.json)
        self.date = datetime(2013, 8, 31)

    def test_as_response(self):
        r = self.interaction.as_response()
        assert isinstance(r, Response)

    def test_as_response_returns_new_instance(self):
        r1 = self.interaction.as_response()
        r2 = self.interaction.as_response()
        assert r1 is not r2

    def test_deserialized_response(self):
        def check_uri(attr):
            # Necessary since PreparedRequests do not have a uri attr
            if attr == 'uri':
                return 'url'
            return attr
        r = self.interaction.as_response()
        for attr in ['status_code', 'url']:
            assert self.response[attr] == decode(getattr(r, attr))

        headers = dict((k, v[0]) for k, v in self.response['headers'].items())
        assert headers == r.headers

        tested_cookie = False
        for cookie in r.cookies:
            cookie_str = "{0}={1}".format(cookie.name, cookie.value)
            assert cookie_str == r.headers['Set-Cookie']
            tested_cookie = True
        assert tested_cookie

        assert self.response['body']['string'] == decode(r.content)
        actual_req = r.request
        expected_req = self.request
        for attr in ['method', 'uri']:
            assert expected_req[attr] == getattr(actual_req, check_uri(attr))

        assert self.request['body']['string'] == decode(actual_req.body)
        headers = dict((k, v[0]) for k, v in expected_req['headers'].items())
        assert headers == actual_req.headers
        assert self.date == self.interaction.recorded_at

    def test_match(self):
        matchers = [lambda x: True, lambda x: False, lambda x: True]
        assert self.interaction.match(matchers) is False
        matchers[1] = lambda x: True
        assert self.interaction.match(matchers) is True

    def test_replace(self):
        self.interaction.replace('123456789abcdef', '<AUTH_TOKEN>')
        self.interaction.replace('cookie_value', '<COOKIE_VALUE>')
        self.interaction.replace('secret_value', '<SECRET_VALUE>')
        self.interaction.replace('foo', '<FOO>')
        self.interaction.replace('http://example.com', '<EXAMPLE_URI>')

        header = self.interaction.data['request']['headers']['Authorization']
        assert header == '<AUTH_TOKEN>'
        header = self.interaction.data['response']['headers']['Set-Cookie']
        assert header == 'cookie_name=<COOKIE_VALUE>'
        body = self.interaction.data['request']['body']['string']
        assert body == 'key=value&key2=<SECRET_VALUE>'
        body = self.interaction.data['response']['body']
        assert body == {'encoding': 'utf-8', 'string': '<FOO>'}
        uri = self.interaction.data['request']['uri']
        assert uri == '<EXAMPLE_URI>/'
        uri = self.interaction.data['response']['url']
        assert uri == '<EXAMPLE_URI>'

    def test_replace_in_headers(self):
        self.interaction.replace_in_headers('123456789abcdef', '<AUTH_TOKEN>')
        self.interaction.replace_in_headers('cookie_value', '<COOKIE_VALUE>')
        header = self.interaction.data['request']['headers']['Authorization']
        assert header == '<AUTH_TOKEN>'
        header = self.interaction.data['response']['headers']['Set-Cookie']
        assert header == 'cookie_name=<COOKIE_VALUE>'

    def test_replace_in_body(self):
        self.interaction.replace_in_body('secret_value', '<SECRET_VALUE>')
        self.interaction.replace_in_body('foo', '<FOO>')
        body = self.interaction.data['request']['body']['string']
        assert body == 'key=value&key2=<SECRET_VALUE>'
        body = self.interaction.data['response']['body']
        assert body == {'encoding': 'utf-8', 'string': '<FOO>'}

    def test_replace_in_uri(self):
        self.interaction.replace_in_uri('http://example.com', '<EXAMPLE_URI>')
        uri = self.interaction.data['request']['uri']
        assert uri == '<EXAMPLE_URI>/'
        uri = self.interaction.data['response']['url']
        assert uri == '<EXAMPLE_URI>'


class TestMockHTTPResponse(unittest.TestCase):
    def setUp(self):
        self.resp = mock_response.MockHTTPResponse(HTTPHeaderDict({
            decode('Header'): decode('value')
        }))

    def test_isclosed(self):
        assert self.resp.isclosed() is False

    def test_is_Message(self):
        assert isinstance(self.resp.msg, email.message.Message)



# -*- coding: utf-8 -*-
from . import matchers, serializers
from .adapter import BetamaxAdapter
from .cassette import Cassette
from .configure import Configuration
from .options import Options


class Betamax(object):

    """This object contains the main API of the request-vcr library.

    This object is entirely a context manager so all you have to do is:

    .. code::

        s = requests.Session()
        with Betamax(s) as vcr:
            vcr.use_cassette('example')
            r = s.get('https://httpbin.org/get')

    Or more concisely, you can do:

    .. code::

        s = requests.Session()
        with Betamax(s).use_cassette('example') as vcr:
            r = s.get('https://httpbin.org/get')

    This object allows for the user to specify the cassette library directory
    and default cassette options.

    .. code::

        s = requests.Session()
        with Betamax(s, cassette_library_dir='tests/cassettes') as vcr:
            vcr.use_cassette('example')
            r = s.get('https://httpbin.org/get')

        with Betamax(s, default_cassette_options={
                're_record_interval': 1000
                }) as vcr:
            vcr.use_cassette('example')
            r = s.get('https://httpbin.org/get')

    """

    def __init__(self, session, cassette_library_dir=None,
                 default_cassette_options={}):
        #: Store the requests.Session object being wrapped.
        self.session = session
        #: Store the session's original adapters.
        self.http_adapters = session.adapters.copy()
        #: Create a new adapter to replace the existing ones
        self.betamax_adapter = BetamaxAdapter(old_adapters=self.http_adapters)
        # We need a configuration instance to make life easier
        self.config = Configuration()
        # Merge the new cassette options with the default ones
        self.config.default_cassette_options.update(
            default_cassette_options or {}
        )

        # If it was passed in, use that instead.
        if cassette_library_dir:
            self.config.cassette_library_dir = cassette_library_dir

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *ex_args):
        self.stop()
        # ex_args comes through as the exception type, exception value and
        # exception traceback. If any of them are not None, we should probably
        # try to raise the exception and not muffle anything.
        if any(ex_args):
            # If you return False, Python will re-raise the exception for you
            return False

    @staticmethod
    def configure():
        """Help to configure the library as a whole.

        .. code::

            with Betamax.configure() as config:
                config.cassette_library_dir = 'tests/cassettes/'
                config.default_cassette_options['match_options'] = [
                    'method', 'uri', 'headers'
                    ]
        """
        return Configuration()

    @property
    def current_cassette(self):
        """Return the cassette that is currently in use.

        :returns: :class:`Cassette <betamax.cassette.Cassette>`
        """
        return self.betamax_adapter.cassette

    @staticmethod
    def register_request_matcher(matcher_class):
        """Register a new request matcher.

        :param matcher_class: (required), this must sub-class
            :class:`BaseMatcher <betamax.matchers.BaseMatcher>`
        """
        matchers.matcher_registry[matcher_class.name] = matcher_class()

    @staticmethod
    def register_serializer(serializer_class):
        """Register a new serializer.

        :param matcher_class: (required), this must sub-class
            :class:`BaseSerializer <betamax.serializers.BaseSerializer>`
        """
        name = serializer_class.name
        serializers.serializer_registry[name] = serializer_class()

    # ❙▸
    def start(self):
        """Start recording or replaying interactions."""
        for k in self.http_adapters:
            self.session.mount(k, self.betamax_adapter)

    # ■
    def stop(self):
        """Stop recording or replaying interactions."""
        # No need to keep the cassette in memory any longer.
        self.betamax_adapter.eject_cassette()
        # On exit, we no longer wish to use our adapter and we want the
        # session to behave normally! Woooo!
        self.betamax_adapter.close()
        for (k, v) in self.http_adapters.items():
            self.session.mount(k, v)

    def use_cassette(self, cassette_name, **kwargs):
        """Tell Betamax which cassette you wish to use for the context.

        :param str cassette_name: relative name, without the serialization
            format, of the cassette you wish Betamax would use
        :param str serialize_with: the format you want Betamax to serialize
            the cassette with
        :param str serialize: DEPRECATED the format you want Betamax to
            serialize the request and response data to and from
        """
        kwargs = Options(kwargs)
        serialize = kwargs['serialize'] or kwargs['serialize_with']
        kwargs['cassette_library_dir'] = self.config.cassette_library_dir

        can_load = Cassette.can_be_loaded(
            self.config.cassette_library_dir,
            cassette_name,
            serialize,
            kwargs['record']
            )

        if can_load:
            self.betamax_adapter.load_cassette(cassette_name, serialize,
                                               kwargs)
        else:
            # If we're not recording or replaying an existing cassette, we
            # should tell the user/developer that there is no cassette, only
            # Zuul
            raise ValueError('Cassette must have a valid name and may not be'
                             ' None.')
        return self



import unittest

from betamax import Betamax
from requests import Session
from requests.auth import HTTPDigestAuth


class TestDigestAuth(unittest.TestCase):
    def test_saves_content_as_gzip(self):
        s = Session()
        cassette_name = 'handles_digest_auth'
        match = ['method', 'uri', 'digest-auth']
        with Betamax(s).use_cassette(cassette_name, match_requests_on=match):
            r = s.get('https://httpbin.org/digest-auth/auth/user/passwd',
                      auth=HTTPDigestAuth('user', 'passwd'))
            assert r.ok
            assert r.history[0].status_code == 401

        s = Session()
        with Betamax(s).use_cassette(cassette_name, match_requests_on=match):
            r = s.get('https://httpbin.org/digest-auth/auth/user/passwd',
                      auth=HTTPDigestAuth('user', 'passwd'))
            assert r.json() is not None



try:
    import unittest.mock as mock
except ImportError:
    import mock

import pytest
import unittest

import requests

import betamax
from betamax.fixtures import pytest as pytest_fixture
from betamax.fixtures import unittest as unittest_fixture


class TestPyTestFixture(unittest.TestCase):
    def setUp(self):
        self.mocked_betamax = mock.MagicMock()
        self.patched_betamax = mock.patch.object(
            betamax.recorder, 'Betamax', return_value=self.mocked_betamax)
        self.patched_betamax.start()

    def tearDown(self):
        self.patched_betamax.stop()

    def test_adds_stop_as_a_finalizer(self):
        # Mock a pytest request object
        request = mock.MagicMock()
        request.cls = request.module = None
        request.function.__name__ = 'test'

        pytest_fixture.betamax_recorder(request)
        assert request.addfinalizer.called is True
        request.addfinalizer.assert_called_once_with(self.mocked_betamax.stop)

    def test_auto_starts_the_recorder(self):
        # Mock a pytest request object
        request = mock.MagicMock()
        request.cls = request.module = None
        request.function.__name__ = 'test'

        pytest_fixture.betamax_recorder(request)
        self.mocked_betamax.start.assert_called_once_with()


class FakeBetamaxTestCase(unittest_fixture.BetamaxTestCase):
    def test_fake(self):
        pass


class TestUnittestFixture(unittest.TestCase):
    def setUp(self):
        self.mocked_betamax = mock.MagicMock()
        self.patched_betamax = mock.patch.object(
            betamax.recorder, 'Betamax', return_value=self.mocked_betamax)
        self.betamax = self.patched_betamax.start()
        self.fixture = FakeBetamaxTestCase(methodName='test_fake')

    def tearDown(self):
        self.patched_betamax.stop()

    def test_setUp(self):
        self.fixture.setUp()

        self.mocked_betamax.use_cassette.assert_called_once_with(
            'FakeBetamaxTestCase.test_fake'
        )
        self.mocked_betamax.start.assert_called_once_with()

    def test_setUp_rejects_arbitrary_session_classes(self):
        self.fixture.SESSION_CLASS = object

        with pytest.raises(AssertionError):
            self.fixture.setUp()

    def test_setUp_accepts_session_subclasses(self):
        class TestSession(requests.Session):
            pass

        self.fixture.SESSION_CLASS = TestSession

        self.fixture.setUp()

        assert self.betamax.called is True
        call_kwargs = self.betamax.call_args[-1]
        assert isinstance(call_kwargs['session'], TestSession)

    def test_tearDown_calls_stop(self):
        recorder = mock.Mock()
        self.fixture.recorder = recorder

        self.fixture.tearDown()

        recorder.stop.assert_called_once_with()



import os.path

import pytest


@pytest.mark.usefixtures('betamax_session')
class TestPyTestFixtures:
    @pytest.fixture(autouse=True)
    def setup(self, request):
        """After test hook to assert everything."""
        def finalizer():
            test_dir = os.path.abspath('.')
            cassette_name = ('tests.integration.test_fixtures.'  # Module name
                             'TestPyTestFixtures.'  # Class name
                             'test_pytest_fixture'  # Test function name
                             '.json')
            file_name = os.path.join(test_dir, 'tests', 'cassettes',
                                     cassette_name)
            assert os.path.exists(file_name) is True

        request.addfinalizer(finalizer)

    def test_pytest_fixture(self, betamax_session):
        """Exercise the fixture itself."""
        resp = betamax_session.get('https://httpbin.org/get')
        assert resp.ok



import collections
import copy
import unittest

from betamax.configure import Configuration
from betamax.cassette import Cassette


class TestConfiguration(unittest.TestCase):
    def setUp(self):
        self.cassette_options = copy.deepcopy(
            Cassette.default_cassette_options
            )
        self.cassette_dir = Configuration.CASSETTE_LIBRARY_DIR

    def tearDown(self):
        Cassette.default_cassette_options = self.cassette_options
        Cassette.hooks = collections.defaultdict(list)
        Configuration.CASSETTE_LIBRARY_DIR = self.cassette_dir

    def test_acts_as_pass_through(self):
        c = Configuration()
        c.default_cassette_options['foo'] = 'bar'
        assert 'foo' in Cassette.default_cassette_options
        assert Cassette.default_cassette_options.get('foo') == 'bar'

    def test_sets_cassette_library(self):
        c = Configuration()
        c.cassette_library_dir = 'foo'
        assert Configuration.CASSETTE_LIBRARY_DIR == 'foo'

    def test_is_a_context_manager(self):
        with Configuration() as c:
            assert isinstance(c, Configuration)

    def test_allows_registration_of_placeholders(self):
        opts = copy.deepcopy(Cassette.default_cassette_options)
        c = Configuration()
        c.define_cassette_placeholder('<TEST>', 'test')

        assert opts != Cassette.default_cassette_options
        placeholders = Cassette.default_cassette_options['placeholders']
        assert placeholders[0]['placeholder'] == '<TEST>'
        assert placeholders[0]['replace'] == 'test'

    def test_registers_pre_record_hooks(self):
        c = Configuration()
        assert Cassette.hooks['before_record'] == []
        c.before_record(callback=lambda: None)
        assert Cassette.hooks['before_record'] != []
        assert len(Cassette.hooks['before_record']) == 1
        assert callable(Cassette.hooks['before_record'][0])

    def test_registers_pre_playback_hooks(self):
        c = Configuration()
        assert Cassette.hooks['before_playback'] == []
        c.before_playback(callback=lambda: None)
        assert Cassette.hooks['before_playback'] != []
        assert len(Cassette.hooks['before_playback']) == 1
        assert callable(Cassette.hooks['before_playback'][0])



import unittest

try:
    from unittest import mock
except ImportError:
    import mock

from betamax.adapter import BetamaxAdapter
from requests.adapters import HTTPAdapter


class TestBetamaxAdapter(unittest.TestCase):
    def setUp(self):
        http_adapter = mock.Mock()
        self.adapters_dict = {'http://': http_adapter}
        self.adapter = BetamaxAdapter(old_adapters=self.adapters_dict)

    def tearDown(self):
        self.adapter.eject_cassette()

    def test_has_http_adatper(self):
        assert self.adapter.http_adapter is not None
        assert isinstance(self.adapter.http_adapter, HTTPAdapter)

    def test_empty_initial_state(self):
        assert self.adapter.cassette is None
        assert self.adapter.cassette_name is None
        assert self.adapter.serialize is None

    def test_load_cassette(self):
        filename = 'test'
        self.adapter.load_cassette(filename, 'json', {
            'record': 'none',
            'cassette_library_dir': 'tests/cassettes/'
        })
        assert self.adapter.cassette is not None
        assert self.adapter.cassette_name == filename



from betamax import Betamax, BetamaxError

from tests.integration.helper import IntegrationHelper


class TestRecordOnce(IntegrationHelper):
    def test_records_new_interaction(self):
        s = self.session
        with Betamax(s).use_cassette('test_record_once') as betamax:
            self.cassette_path = betamax.current_cassette.cassette_path
            assert betamax.current_cassette.is_empty() is True
            r = s.get('http://httpbin.org/get')
            assert r.status_code == 200
            assert betamax.current_cassette.is_empty() is True
            assert betamax.current_cassette.interactions != []

    def test_replays_response_from_cassette(self):
        s = self.session
        with Betamax(s).use_cassette('test_replays_response') as betamax:
            self.cassette_path = betamax.current_cassette.cassette_path
            assert betamax.current_cassette.is_empty() is True
            r0 = s.get('http://httpbin.org/get')
            assert r0.status_code == 200
            assert betamax.current_cassette.interactions != []
            assert len(betamax.current_cassette.interactions) == 1
            r1 = s.get('http://httpbin.org/get')
            assert len(betamax.current_cassette.interactions) == 2
            assert r1.status_code == 200
            r0_headers = r0.headers.copy()
            r0_headers.pop('Date')
            r0_headers.pop('Age', None)
            r1_headers = r1.headers.copy()
            r1_headers.pop('Date')
            r1_headers.pop('Age', None)
            # NOTE(sigmavirus24): This fails if the second request is
            # technically a second later. Ignoring the Date headers allows
            # this test to succeed.
            assert r0_headers == r1_headers
            assert r0.content == r1.content


class TestRecordNone(IntegrationHelper):
    def test_raises_exception_when_no_interactions_present(self):
        s = self.session
        with Betamax(s) as betamax:
            betamax.use_cassette('test', record='none')
            self.cassette_created = False
            assert betamax.current_cassette is not None
            self.assertRaises(BetamaxError, s.get, 'http://httpbin.org/get')

    def test_record_none_does_not_create_cassettes(self):
        s = self.session
        with Betamax(s) as betamax:
            self.assertRaises(ValueError, betamax.use_cassette,
                              'test_record_none', record='none')
        self.cassette_created = False


class TestRecordNewEpisodes(IntegrationHelper):
    def setUp(self):
        super(TestRecordNewEpisodes, self).setUp()
        with Betamax(self.session).use_cassette('test_record_new'):
            self.session.get('http://httpbin.org/get')
            self.session.get('http://httpbin.org/redirect/2')

    def test_records_new_events_with_existing_cassette(self):
        s = self.session
        opts = {'record': 'new_episodes'}
        with Betamax(s).use_cassette('test_record_new', **opts) as betamax:
            cassette = betamax.current_cassette
            self.cassette_path = cassette.cassette_path
            assert cassette.interactions != []
            assert len(cassette.interactions) == 4
            assert cassette.is_empty() is False
            s.get('https://httpbin.org/get')
            assert len(cassette.interactions) == 5

        with Betamax(s).use_cassette('test_record_new') as betamax:
            cassette = betamax.current_cassette
            assert len(cassette.interactions) == 5
            r = s.get('https://httpbin.org/get')
            assert r.status_code == 200


class TestRecordNewEpisodesCreatesCassettes(IntegrationHelper):
    def test_creates_new_cassettes(self):
        recorder = Betamax(self.session)
        opts = {'record': 'new_episodes'}
        cassette_name = 'test_record_new_makes_new_cassettes'
        with recorder.use_cassette(cassette_name, **opts) as betamax:
            self.cassette_path = betamax.current_cassette.cassette_path
            self.session.get('https://httpbin.org/get')


class TestRecordAll(IntegrationHelper):
    def setUp(self):
        super(TestRecordAll, self).setUp()
        with Betamax(self.session).use_cassette('test_record_all'):
            self.session.get('http://httpbin.org/get')
            self.session.get('http://httpbin.org/redirect/2')
            self.session.get('http://httpbin.org/get')

    def test_records_new_interactions(self):
        s = self.session
        opts = {'record': 'all'}
        with Betamax(s).use_cassette('test_record_all', **opts) as betamax:
            cassette = betamax.current_cassette
            self.cassette_path = cassette.cassette_path
            assert cassette.interactions != []
            assert len(cassette.interactions) == 5
            assert cassette.is_empty() is False
            s.post('http://httpbin.org/post', data={'foo': 'bar'})
            assert len(cassette.interactions) == 6

        with Betamax(s).use_cassette('test_record_all') as betamax:
            assert len(betamax.current_cassette.interactions) == 6

    def test_replaces_old_interactions(self):
        s = self.session
        opts = {'record': 'all'}
        with Betamax(s).use_cassette('test_record_all', **opts) as betamax:
            cassette = betamax.current_cassette
            self.cassette_path = cassette.cassette_path
            assert cassette.interactions != []
            assert len(cassette.interactions) == 5
            assert cassette.is_empty() is False
            s.get('http://httpbin.org/get')
            assert len(cassette.interactions) == 5



import unittest

from betamax import matchers, serializers
from betamax.adapter import BetamaxAdapter
from betamax.cassette import cassette
from betamax.recorder import Betamax
from requests import Session
from requests.adapters import HTTPAdapter


class TestBetamax(unittest.TestCase):
    def setUp(self):
        self.session = Session()
        self.vcr = Betamax(self.session)

    def test_initialization_does_not_alter_the_session(self):
        for v in self.session.adapters.values():
            assert not isinstance(v, BetamaxAdapter)
            assert isinstance(v, HTTPAdapter)

    def test_initialization_converts_placeholders(self):
        placeholders = [{'placeholder': '<FOO>', 'replace': 'replace-with'}]
        default_cassette_options = {'placeholders': placeholders}
        self.vcr = Betamax(self.session,
                           default_cassette_options=default_cassette_options)
        assert self.vcr.config.default_cassette_options['placeholders'] == [{
            'placeholder': '<FOO>',
            'replace': 'replace-with',
        }]

    def test_entering_context_alters_adapters(self):
        with self.vcr:
            for v in self.session.adapters.values():
                assert isinstance(v, BetamaxAdapter)

    def test_exiting_resets_the_adapters(self):
        with self.vcr:
            pass
        for v in self.session.adapters.values():
            assert not isinstance(v, BetamaxAdapter)

    def test_current_cassette(self):
        assert self.vcr.current_cassette is None
        self.vcr.use_cassette('test')
        assert isinstance(self.vcr.current_cassette, cassette.Cassette)

    def test_use_cassette_returns_cassette_object(self):
        assert self.vcr.use_cassette('test') is self.vcr

    def test_register_request_matcher(self):
        class FakeMatcher(object):
            name = 'fake_matcher'

        Betamax.register_request_matcher(FakeMatcher)
        assert 'fake_matcher' in matchers.matcher_registry
        assert isinstance(matchers.matcher_registry['fake_matcher'],
                          FakeMatcher)

    def test_register_serializer(self):
        class FakeSerializer(object):
            name = 'fake_serializer'

        Betamax.register_serializer(FakeSerializer)
        assert 'fake_serializer' in serializers.serializer_registry
        assert isinstance(serializers.serializer_registry['fake_serializer'],
                          FakeSerializer)

    def test_stores_the_session_instance(self):
        assert self.session is self.vcr.session

    def test_use_cassette_passes_along_placeholders(self):
        placeholders = [{'placeholder': '<FOO>', 'replace': 'replace-with'}]
        self.vcr.use_cassette('test', placeholders=placeholders)
        assert self.vcr.current_cassette.placeholders == [
            cassette.Placeholder.from_dict(p) for p in placeholders
        ]



from betamax import Betamax, BetamaxError
from requests import Session

import unittest


class TestReplays(unittest.TestCase):
    def setUp(self):
        self.session = Session()

    def test_replays_response_on_right_order(self):
        s = self.session
        opts = {'record': 'none'}
        with Betamax(s).use_cassette('test_replays_response_on_right_order', **opts) as betamax:
            self.cassette_path = betamax.current_cassette.cassette_path
            r0 = s.get('http://httpbin.org/get')
            r1 = s.get('http://httpbin.org/get')
            r0_found = (b'72.160.214.132' in r0.content)
            assert r0_found == True
            r1_found = (b'72.160.214.133' in r1.content)
            assert r1_found == True



# -*- coding: utf-8 -*-
from .base import BaseMatcher
from betamax.util import deserialize_prepared_request


class BodyMatcher(BaseMatcher):
    # Matches based on the body of the request
    name = 'body'

    def match(self, request, recorded_request):
        recorded_request = deserialize_prepared_request(recorded_request)

        if request.body:
            if isinstance(recorded_request.body, type(request.body)):
                request_body = request.body
            else:
                request_body = request.body.encode('utf-8')
        else:
            request_body = b''

        return recorded_request.body == request_body



# -*- coding: utf-8 -*-
from .base import BaseMatcher


class HeadersMatcher(BaseMatcher):
    # Matches based on the headers of the request
    name = 'headers'

    def match(self, request, recorded_request):
        return dict(request.headers) == self.flatten_headers(recorded_request)

    def flatten_headers(self, request):
        from betamax.util import from_list
        headers = request['headers'].items()
        return dict((k, from_list(v)) for (k, v) in headers)



import betamax
import copy
from .helper import IntegrationHelper


class TestBackwardsCompatibleSerialization(IntegrationHelper):
    def setUp(self):
        super(TestBackwardsCompatibleSerialization, self).setUp()
        self.cassette_created = False
        opts = betamax.cassette.Cassette.default_cassette_options
        self.original_defaults = copy.deepcopy(opts)

        with betamax.Betamax.configure() as config:
            config.define_cassette_placeholder('<FOO>', 'nothing to replace')

    def tearDown(self):
        super(TestBackwardsCompatibleSerialization, self).setUp()
        Cassette = betamax.cassette.Cassette
        Cassette.default_cassette_options = self.original_defaults

    def test_can_deserialize_an_old_cassette(self):
        with betamax.Betamax(self.session).use_cassette('GitHub_emojis') as b:
            assert b.current_cassette is not None
            cassette = b.current_cassette
            assert len(cassette.interactions) > -1

    def test_matches_old_request_data(self):
        with betamax.Betamax(self.session).use_cassette('GitHub_emojis'):
            r = self.session.get('https://api.github.com/emojis')
            assert r is not None

    def tests_populates_correct_fields_with_missing_data(self):
        with betamax.Betamax(self.session).use_cassette('GitHub_emojis'):
            r = self.session.get('https://api.github.com/emojis')
            assert r.reason == 'OK'
            assert r.status_code == 200

    def tests_deserializes_old_cassette_headers(self):
        with betamax.Betamax(self.session).use_cassette('GitHub_emojis') as b:
            self.session.get('https://api.github.com/emojis')
            interaction = b.current_cassette.interactions[0].data
            header = interaction['request']['headers']['Accept']
            assert not isinstance(header, list)



from .mock_response import MockHTTPResponse
from datetime import datetime
from requests.models import PreparedRequest, Response
from requests.packages.urllib3 import HTTPResponse
from requests.structures import CaseInsensitiveDict
from requests.status_codes import _codes
from requests.cookies import RequestsCookieJar

try:
    from requests.packages.urllib3._collections import HTTPHeaderDict
except ImportError:
    from .headers import HTTPHeaderDict

import base64
import io
import sys


def coerce_content(content, encoding=None):
    if hasattr(content, 'decode'):
        content = content.decode(encoding or 'utf-8', 'replace')
    return content


def body_io(string, encoding=None):
    if hasattr(string, 'encode'):
        string = string.encode(encoding or 'utf-8')
    return io.BytesIO(string)


def from_list(value):
    if isinstance(value, list):
        return value[0]
    return value


def add_body(r, preserve_exact_body_bytes, body_dict):
    """Simple function which takes a response or request and coerces the body.

    This function adds either ``'string'`` or ``'base64_string'`` to
    ``body_dict``. If ``preserve_exact_body_bytes`` is ``True`` then it
    encodes the body as a base64 string and saves it like that. Otherwise,
    it saves the plain string.

    :param r: This is either a PreparedRequest instance or a Response
        instance.
    :param preserve_exact_body_bytes bool: Either True or False.
    :param body_dict dict: A dictionary already containing the encoding to be
        used.
    """
    body = getattr(r, 'raw', getattr(r, 'body', None))
    if hasattr(body, 'read'):
        body = body.read()

    if not body:
        body = ''

    if (preserve_exact_body_bytes or
            'gzip' in r.headers.get('Content-Encoding', '')):
        if sys.version_info >= (3, 0) and hasattr(body, 'encode'):
            body = body.encode(body_dict['encoding'] or 'utf-8')

        body_dict['base64_string'] = base64.b64encode(body).decode()
    else:
        body_dict['string'] = coerce_content(body, body_dict['encoding'])


def serialize_prepared_request(request, preserve_exact_body_bytes):
    headers = request.headers
    body = {'encoding': 'utf-8'}
    add_body(request, preserve_exact_body_bytes, body)
    return {
        'body': body,
        'headers': dict(
            (coerce_content(k, 'utf-8'), [v]) for (k, v) in headers.items()
        ),
        'method': request.method,
        'uri': request.url,
    }


def deserialize_prepared_request(serialized):
    p = PreparedRequest()
    p._cookies = RequestsCookieJar()
    body = serialized['body']
    if isinstance(body, dict):
        original_body = body.get('string')
        p.body = original_body or base64.b64decode(
            body.get('base64_string', '').encode())
    else:
        p.body = body
    h = [(k, from_list(v)) for k, v in serialized['headers'].items()]
    p.headers = CaseInsensitiveDict(h)
    p.method = serialized['method']
    p.url = serialized['uri']
    return p


def serialize_response(response, preserve_exact_body_bytes):
    body = {'encoding': response.encoding}
    add_body(response, preserve_exact_body_bytes, body)
    header_map = HTTPHeaderDict(response.raw.headers)
    headers = {}
    for header_name in header_map.keys():
        headers[header_name] = header_map.getlist(header_name)

    return {
        'body': body,
        'headers': headers,
        'status': {'code': response.status_code, 'message': response.reason},
        'url': response.url,
    }


def deserialize_response(serialized):
    r = Response()
    r.encoding = serialized['body']['encoding']
    header_dict = HTTPHeaderDict()

    for header_name, header_list in serialized['headers'].items():
        if isinstance(header_list, list):
            for header_value in header_list:
                header_dict.add(header_name, header_value)
        else:
            header_dict.add(header_name, header_list)
    r.headers = CaseInsensitiveDict(header_dict)

    r.url = serialized.get('url', '')
    if 'status' in serialized:
        r.status_code = serialized['status']['code']
        r.reason = serialized['status']['message']
    else:
        r.status_code = serialized['status_code']
        r.reason = _codes[r.status_code][0].upper()
    add_urllib3_response(serialized, r, header_dict)
    return r


def add_urllib3_response(serialized, response, headers):
    if 'base64_string' in serialized['body']:
        body = io.BytesIO(
            base64.b64decode(serialized['body']['base64_string'].encode())
        )
    else:
        body = body_io(**serialized['body'])

    h = HTTPResponse(
        body,
        status=response.status_code,
        reason=response.reason,
        headers=headers,
        preload_content=False,
        original_response=MockHTTPResponse(headers)
    )
    # NOTE(sigmavirus24):
    # urllib3 updated it's chunked encoding handling which breaks on recorded
    # responses. Since a recorded response cannot be streamed appropriately
    # for this handling to work, we can preserve the integrity of the data in
    # the response by forcing the chunked attribute to always be False.
    # This isn't pretty, but it is much better than munging a response.
    h.chunked = False
    response.raw = h


def timestamp():
    stamp = datetime.utcnow().isoformat()
    try:
        i = stamp.rindex('.')
    except ValueError:
        return stamp
    else:
        return stamp[:i]


_SENTINEL = object()


def _option_from(option, kwargs, defaults):
    value = kwargs.get(option, _SENTINEL)
    if value is _SENTINEL:
        value = defaults.get(option)
    return value



from .cassette import Cassette


class Configuration(object):

    """This object acts as a proxy to configure different parts of Betamax.

    You should only ever encounter this object when configuring the library as
    a whole. For example:

    .. code::

        with Betamax.configure() as config:
            config.cassette_library_dir = 'tests/cassettes/'
            config.default_cassette_options['record_mode'] = 'once'
            config.default_cassette_options['match_requests_on'] = ['uri']
            config.define_cassette_placeholder('<URI>', 'http://httpbin.org')
            config.preserve_exact_body_bytes = True

    """

    CASSETTE_LIBRARY_DIR = 'vcr/cassettes'

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def __setattr__(self, prop, value):
        if prop == 'preserve_exact_body_bytes':
            self.default_cassette_options[prop] = True
        else:
            super(Configuration, self).__setattr__(prop, value)

    def before_playback(self, tag=None, callback=None):
        """Register a function to call before playing back an interaction.

        Example usage:

        .. code-block:: python

            def before_playback(interaction, cassette):
                pass

            with Betamax.configure() as config:
                config.before_playback(callback=before_playback)

        :param str tag:
            Limits the interactions passed to the function based on the
            interaction's tag (currently unsupported).
        :param callable callback:
            The function which either accepts just an interaction or an
            interaction and a cassette and mutates the interaction before
            returning.
        """
        Cassette.hooks['before_playback'].append(callback)

    def before_record(self, tag=None, callback=None):
        """Register a function to call before recording an interaction.

        Example usage:

        .. code-block:: python

            def before_record(interaction, cassette):
                pass

            with Betamax.configure() as config:
                config.before_record(callback=before_record)

        :param str tag:
            Limits the interactions passed to the function based on the
            interaction's tag (currently unsupported).
        :param callable callback:
            The function which either accepts just an interaction or an
            interaction and a cassette and mutates the interaction before
            returning.
        """
        Cassette.hooks['before_record'].append(callback)

    @property
    def cassette_library_dir(self):
        """Retrieve and set the directory to store the cassettes in."""
        return Configuration.CASSETTE_LIBRARY_DIR

    @cassette_library_dir.setter
    def cassette_library_dir(self, value):
        Configuration.CASSETTE_LIBRARY_DIR = value

    @property
    def default_cassette_options(self):
        """Retrieve and set the default cassette options.

        The options include:

        - ``match_requests_on``
        - ``placeholders``
        - ``re_record_interval``
        - ``record_mode``
        - ``preserve_exact_body_bytes``

        Other options will be ignored.
        """
        return Cassette.default_cassette_options

    @default_cassette_options.setter
    def default_cassette_options(self, value):
        Cassette.default_cassette_options = value

    def define_cassette_placeholder(self, placeholder, replace):
        """Define a placeholder value for some text.

        This also will replace the placeholder text with the text you wish it
        to use when replaying interactions from cassettes.

        :param str placeholder: (required), text to be used as a placeholder
        :param str replace: (required), text to be replaced or replacing the
            placeholder
        """
        self.default_cassette_options['placeholders'].append({
            'placeholder': placeholder,
            'replace': replace,
        })



import unittest

from betamax import Betamax
from requests import Session


class TestReplayInteractionMultipleTimes(unittest.TestCase):
    """
    Test that an Interaction can be replayed multiple times within the same
    betamax session.
    """
    def test_replay_interaction_more_than_once(self):
        s = Session()

        with Betamax(s).use_cassette('replay_multiple_times', record='once',
                                     allow_playback_repeats=True):
            for k in range(1, 5):
                r = s.get('http://httpbin.org/stream/3', stream=True)
                assert r.raw.read(1028), "Stream already consumed. Try: %d" % k



import os
import re
import sys
from setuptools import setup, find_packages

packages = find_packages(exclude=['tests', 'tests.integration'])
requires = ['requests >= 2.0']

__version__ = ''
with open('betamax/__init__.py', 'r') as fd:
    reg = re.compile(r'__version__ = [\'"]([^\'"]*)[\'"]')
    for line in fd:
        m = reg.match(line)
        if m:
            __version__ = m.group(1)
            break

if not __version__:
    raise RuntimeError('Cannot find version information')

if sys.argv[-1] in ['submit', 'publish']:
    os.system("python setup.py sdist bdist_wheel upload")
    sys.exit()


def data_for(filename):
    with open(filename) as fd:
        content = fd.read()
    return content

setup(
    name="betamax",
    version=__version__,
    description="A VCR imitation for python-requests",
    long_description="\n\n".join([data_for("README.rst"),
                                  data_for("HISTORY.rst")]),
    license="Apache 2.0",
    author="Ian Cordasco",
    author_email="graffatcolmingov@gmail.com",
    url="https://github.com/sigmavirus24/betamax",
    packages=packages,
    package_data={'': ['LICENSE', 'AUTHORS.rst']},
    include_package_data=True,
    install_requires=requires,
    entry_points={
        'pytest11': ['pytest-betamax = betamax.fixtures.pytest']
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
    ]
)



try:
    from unittest import mock
except ImportError:
    import mock

import betamax
from betamax.decorator import use_cassette


@mock.patch('betamax.recorder.Betamax', autospec=True)
def test_wraps_session(Betamax):
    # This needs to be a magic mock so it will mock __exit__
    recorder = mock.MagicMock(spec=betamax.Betamax)
    recorder.use_cassette.return_value = recorder
    Betamax.return_value = recorder

    @use_cassette('foo', cassette_library_dir='fizbarbogus')
    def _test(session):
        pass

    _test()
    Betamax.assert_called_once_with(
        session=mock.ANY,
        cassette_library_dir='fizbarbogus',
        default_cassette_options={}
    )
    recorder.use_cassette.assert_called_once_with('foo')


@mock.patch('betamax.recorder.Betamax', autospec=True)
@mock.patch('requests.Session')
def test_creates_a_new_session(Session, Betamax):
    @use_cassette('foo', cassette_library_dir='dir')
    def _test(session):
        pass

    _test()

    assert Session.call_count == 1



# -*- coding: utf-8 -*-


class BaseMatcher(object):

    """
    Base class that ensures sub-classes that implement custom matchers can be
    registered and have the only method that is required.

    Usage:

    .. code-block:: python

        from betamax import Betamax, BaseMatcher

        class MyMatcher(BaseMatcher):
            name = 'my'

            def match(self, request, recorded_request):
                # My fancy matching algorithm

        Betamax.register_request_matcher(MyMatcher)

    The last line is absolutely necessary.

    The `match` method will be given a `requests.PreparedRequest` object and a
    dictionary. The dictionary always has the following keys:

    - url
    - method
    - body
    - headers

    """

    name = None

    def __init__(self):
        if not self.name:
            raise ValueError('Matchers require names')
        self.on_init()

    def on_init(self):
        """Method to implement if you wish something to happen in ``__init__``.

        The return value is not checked and this is called at the end of
        ``__init__``. It is meant to provide the matcher author a way to
        perform things during initialization of the instance that would
        otherwise require them to override ``BaseMatcher.__init__``.
        """
        return None

    def match(self, request, recorded_request):
        """A method that must be implemented by the user.

        :param PreparedRequest request: A requests PreparedRequest object
        :param dict recorded_request: A dictionary containing the serialized
            request in the cassette
        :returns bool: True if they match else False
        """
        raise NotImplementedError('The match method must be implemented on'
                                  ' %s' % self.__class__.__name__)



import unittest
from itertools import permutations
from betamax.options import Options, validate_record, validate_matchers


class TestValidators(unittest.TestCase):
    def test_validate_record(self):
        for mode in ['once', 'none', 'all', 'new_episodes']:
            assert validate_record(mode) is True

    def test_validate_matchers(self):
        matchers = ['method', 'uri', 'query', 'host', 'body']
        for i in range(1, len(matchers)):
            for l in permutations(matchers, i):
                assert validate_matchers(l) is True

        matchers.append('foobar')
        assert validate_matchers(matchers) is False


class TestOptions(unittest.TestCase):
    def setUp(self):
        self.data = {
            're_record_interval': 10000,
            'match_requests_on': ['method'],
            'serialize': 'json'
        }
        self.options = Options(self.data)

    def test_data_is_valid(self):
        for key in self.data:
            assert key in self.options

    def test_invalid_data_is_removed(self):
        data = self.data.copy()
        data['fake'] = 'value'
        options = Options(data)

        for key in self.data:
            assert key in options

        assert 'fake' not in options

    def test_values_are_validated(self):
        assert self.options['re_record_interval'] == 10000
        assert self.options['match_requests_on'] == ['method']

        data = self.data.copy()
        data['match_requests_on'] = ['foo', 'bar', 'bogus']
        options = Options(data)
        assert options['match_requests_on'] == ['method', 'uri']



class BetamaxError(Exception):
    def __init__(self, message):
        super(BetamaxError, self).__init__(message)

    def __repr__(self):
        return 'BetamaxError("%s")' % self.message


class MissingDirectoryError(BetamaxError):
    pass



# -*- coding: utf-8 -*-
from .base import BaseMatcher
from .query import QueryMatcher
from requests.compat import urlparse


class URIMatcher(BaseMatcher):
    # Matches based on the uri of the request
    name = 'uri'

    def on_init(self):
        # Get something we can use to match query strings with
        self.query_matcher = QueryMatcher().match

    def match(self, request, recorded_request):
        queries_match = self.query_matcher(request, recorded_request)
        request_url, recorded_url = request.url, recorded_request['uri']
        return self.all_equal(request_url, recorded_url) and queries_match

    def parse(self, uri):
        parsed = urlparse(uri)
        return {
            'scheme': parsed.scheme,
            'netloc': parsed.netloc,
            'path': parsed.path,
            'fragment': parsed.fragment
            }

    def all_equal(self, new_uri, recorded_uri):
        new_parsed = self.parse(new_uri)
        recorded_parsed = self.parse(recorded_uri)
        return (new_parsed == recorded_parsed)



# -*- coding: utf-8 -*-
import sys

from .base import BaseMatcher

try:
    from urlparse import parse_qs, urlparse
except ImportError:
    from urllib.parse import parse_qs, urlparse


isPY2 = (2, 6) <= sys.version_info < (3, 0)


class QueryMatcher(BaseMatcher):
    # Matches based on the query of the request
    name = 'query'

    def to_dict(self, query):
        """Turn the query string into a dictionary."""
        return parse_qs(
            query or '',  # Protect against None
            keep_blank_values=True,
        )

    def match(self, request, recorded_request):
        request_query_dict = self.to_dict(urlparse(request.url).query)
        recorded_query = urlparse(recorded_request['uri']).query
        if recorded_query and isPY2:
            # NOTE(sigmavirus24): If we're on Python 2, the request.url will
            # be str/bytes and the recorded_request['uri'] will be unicode.
            # For the comparison to work for high unicode characters, we need
            # to encode the recorded query string before parsing it. See also
            # GitHub bug #43.
            recorded_query = recorded_query.encode('utf-8')
        recorded_query_dict = self.to_dict(recorded_query)
        return request_query_dict == recorded_query_dict



# -*- coding: utf-8 -*-
NOT_IMPLEMENTED_ERROR_MSG = ('This method must be implemented by classes'
                             ' inheriting from BaseSerializer.')


class BaseSerializer(object):

    """
    Base Serializer class that provides an interface for other serializers.

    Usage:

    .. code-block:: python

        from betamax import Betamax, BaseSerializer


        class MySerializer(BaseSerializer):
            name = 'my'

            @staticmethod
            def generate_cassette_name(cassette_library_dir, cassette_name):
                # Generate a string that will give the relative path of a
                # cassette

            def serialize(self, cassette_data):
                # Take a dictionary and convert it to whatever

            def deserialize(self):
                # Uses a cassette file to return a dictionary with the
                # cassette information

        Betamax.register_serializer(MySerializer)

    The last line is absolutely necessary.

    """

    name = None

    @staticmethod
    def generate_cassette_name(cassette_library_dir, cassette_name):
        raise NotImplementedError(NOT_IMPLEMENTED_ERROR_MSG)

    def __init__(self):
        if not self.name:
            raise ValueError("Serializer's name attribute must be a string"
                             " value, not None.")

        self.on_init()

    def on_init(self):
        """Method to implement if you wish something to happen in ``__init__``.

        The return value is not checked and this is called at the end of
        ``__init__``. It is meant to provide the matcher author a way to
        perform things during initialization of the instance that would
        otherwise require them to override ``BaseSerializer.__init__``.
        """
        return None

    def serialize(self, cassette_data):
        """A method that must be implemented by the Serializer author.

        :param dict cassette_data: A dictionary with two keys:
            ``http_interactions``, ``recorded_with``.
        :returns: Serialized data as a string.
        """
        raise NotImplementedError(NOT_IMPLEMENTED_ERROR_MSG)

    def deserialize(self, cassette_data):
        """A method that must be implemented by the Serializer author.

        The return value is extremely important. If it is not empty, the
        dictionary returned must have the following structure::

            {
                'http_interactions': [{
                    # Interaction
                },
                {
                    # Interaction
                }],
                'recorded_with': 'name of recorder'
            }

        :params str cassette_data: The data serialized as a string which needs
            to be deserialized.
        :returns: dictionary
        """
        raise NotImplementedError(NOT_IMPLEMENTED_ERROR_MSG)
